/*
 * PeerRegistry.java
 *
 * Created on April 1, 2005, 9:09 AM
 */

package jxl.peermi;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * <p>
 * PeerRegistry is the universal lookup point for PeerRemote objects. Once
 * activated, PeerRegistries use jxta multicast to locate one another. After
 * that all lookup operations are forwarded to the peer registries and bound
 * remote objects are returned.
 * </p>
 * 
 * <b>MODIFIED:</b> Since JXTA/P2PSockets seem to have trouble finding
 * neighbour hosts, we'll try to use local multicast instead. /Ola
 * 
 * 
 * @author Alex Lynch
 * @author olale
 * @version 1.1
 */

public final class PeerRegistry extends PeerRemoteObject implements PeerRegistryInterface {

	/**
     * 
     */
	private static final long		serialVersionUID = 1L;

	// start STATIC
	private static PeerRegistry local;

	private static Logger			logger		   = Logger.getLogger(PeerRegistry.class);

	/**
     * Get singleton instance of PeerRegistry.
     */
	public static PeerRegistry getLocalInstance() throws RemoteException {
		if (local == null) {
			local = new PeerRegistry();
		}
		return local;
	}

	// end STATIC

	// start MY FIELDS
	private final HashMap<String, Set<DiscoveryListener>> listenerMap   = new HashMap<String, Set<DiscoveryListener>>();

	private final HashMap<String, PeerRemote>			 regMap		= new HashMap<String, PeerRemote>();

	private final PeerHashSet<PeerRegistryInterface>	  otherRegs	 = new PeerHashSet<PeerRegistryInterface>();

	private MulticastSocket							   multicastSocket;

	private int										   packetSize	= 50000;

	private InetAddress								   group;

	private int										   multicastPort = 6789;

	private boolean									   active		= true;

	private PeerRegistryInterface stub = (PeerRegistryInterface) toStub(this);
	// end MY FIELDS;

	private PeerRegistry() throws RemoteException {
		try {
			otherRegs.add(stub);// I'm in my list
			initMulticast();
			// Start local multicast listener
			Runnable multicastReceiver = new Runnable() {
				public void run() {
					while (isActive()) {
						try {
							addRegistry(receiveRegistry());
						} catch (RemoteException e) {
							logException(e);
						}
					}

				}
			};
			new Thread(multicastReceiver, "multicast receiver").start();
			// Start multicast sender
			Runnable multicastSender = new Runnable() {
				public void run() {
					while (isActive()) {
						sendRegistry();
						logger.debug("sending registry over multicast");
						try {
							Thread.sleep(15000);
						} catch (InterruptedException e) {
							logException(e);
						}
					}

				}
			};
			new Thread(multicastSender, "multicast sender").start();

		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	public void initMulticast() {
		try {
			group = InetAddress.getByName("236.0.0.0");
			multicastSocket = new MulticastSocket(multicastPort);
			multicastSocket.joinGroup(group);
		} catch (UnknownHostException e) {
			logException(e);
		} catch (IOException e) {
			logException(e);
		}
	}

	public PeerRegistryInterface receiveRegistry() {
		try {
			byte[] recvBuf = new byte[packetSize];
			DatagramPacket packet = new DatagramPacket(recvBuf, recvBuf.length);
			multicastSocket.receive(packet);
			ObjectInputStream is = new ObjectInputStream(new BufferedInputStream(new ByteArrayInputStream(recvBuf)));
			Object o = is.readObject();
			is.close();
			return (PeerRegistryInterface) o;
		} catch (IOException e) {
			logException(e);
		} catch (ClassNotFoundException e) {
			logException(e);
		}
		return null;
	}

	public void sendRegistry() {
		try {
			ByteArrayOutputStream byteStream = new ByteArrayOutputStream(packetSize);
			ObjectOutputStream os = new ObjectOutputStream(new BufferedOutputStream(byteStream));
			os.flush();
			os.writeObject(stub);
			os.flush();
			// retrieves byte array
			byte[] sendBuf = byteStream.toByteArray();
			int currentPacketSize = sendBuf.length;
			logger.debug("Sending object of size "+currentPacketSize+" bytes");
			DatagramPacket packet = new DatagramPacket(sendBuf, currentPacketSize, group, multicastPort);
			multicastSocket.send(packet);
			os.close();
		} catch (UnknownHostException e) {
			logException(e);
		} catch (IOException e) {
			logException(e);
		}
	}

	private void logException(Exception e) {
		StringWriter str = new StringWriter();
		e.printStackTrace(new PrintWriter(str));
		logger.error(str);
	}

	private void fireDiscoveryEvent(final PeerRegistryInterface reg) {
		logger.debug("Discovery event triggered");
		new Thread() {
			public void run() {
				try {
					for (String name : listenerMap.keySet()) {
						PeerRemote remote = reg.lookup(name);
						if (remote != null) {
							for (DiscoveryListener list : listenerMap.get(name)) {
								list.discovery(remote);
							}
						} else {
							logger.debug("PeerRegistry at "+reg.getHostName()+" didn't contain "+name);
						}
					}
				} catch (Exception e) {
					logException(e);
				}
			}
		}.start();
	}

	public void addDiscoveryListener(String binding, DiscoveryListener listener) {
		getListenersFor(binding).add(listener);
	}

	private Set<DiscoveryListener> getListenersFor(String name) {
		Set<DiscoveryListener> ret = listenerMap.get(name);
		if (ret == null) {
			ret = new HashSet<DiscoveryListener>();
			listenerMap.put(name, ret);
		}
		return ret;
	}

	public void removeDiscoveryListener(String binding, DiscoveryListener listener) {
		getListenersFor(binding).remove(listener);
	}

	/**
     * Remote method. Should not be called by clients.
     */
	public synchronized void addRegistry(PeerRegistryInterface reg) throws RemoteException {		
		logger.debug("Registry found at " + reg.getHostName());
		if (otherRegs.add(reg)) {
			fireDiscoveryEvent(reg);
		}
		
	}

	/**
     * If <CODE>name</CODE> is not already bound, bind <CODE>p</CODE> to
     * <CODE>name</CODE>. If <CODE>p</CODE> is an instance of
     * PeerRemoteObject the result of <CODE>toStub()</CODE> is bound to this
     * registry, else <CODE>p</CODE> is asumed to be a remote stub, and it is
     * bound itself.
     * <P>
     * If <CODE>name</CODE> was bound return false and do not bind <CODE>p</CODE>.
     * </P>
     * 
     * @param p
     * @param name
     * @throws java.rmi.RemoteException
     * @return true if <CODE>p</CODE> was bound, false otherwise.
     */
	public boolean bind(PeerRemoteObject p, String name) throws RemoteException {
		if (regMap.containsKey(name)) {
			return false;
		} else {
			regMap.put(name, p);
			logger.debug("Registering object with name "+name);
			return true;
		}
	}

	/**
     * Remote method. This can be used to lookup only the local instance of a
     * PeerRemote object. Most clients will use
     * {@link jxl.peermi.PeerRegistry#requestPeerManager(String, Class) }
     * instead.
     */
	public PeerRemote lookup(String name) throws RemoteException {
		return regMap.get(name);
	}

	/**
     * Return PeerManager for binding <code>name</code> for interfact type
     * <code>ifaceType</code>.
     * 
     * @param name
     *            Binding to lookup.
     * @throws java.rmi.RemoteException
     * @return A PeerHashSet containing all of the remote references.
     */
	public <T extends PeerRemote> PeerManager<T> requestPeerManager(String name, Class<T> ifaceType) throws RemoteException {
		PeerManager<T> ret = new PeerManager<T>(ifaceType);
		addDiscoveryListener(name, ret);
		return ret;
	}

	public String getHostName() throws RemoteException {
		InetAddress localHost = null;
		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			throw new RemoteException(e.getMessage());
		}
		return (localHost != null ? localHost.getHostName() : "unknown host address");
	}

	/**
     * @return Returns the active.
     */
	public boolean isActive() {
		return active;
	}

	/**
     * @param active
     *            The active to set.
     */
	public void setActive(boolean active) {
		this.active = active;
	}

}
